-- vram	256k x 8 - 
-- this version is optimized for 8 bits ie it stores 4 bytes per integer
-- using the memin and memout procedures/functions
library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;

ENTITY pulse_widths_vram IS
    GENERIC (
        tlow: TIME := 0 ns;
        thigh: TIME := 0 ns);
    PORT (
        in0 : IN std_logic;
        out0: OUT std_logic);
END pulse_widths_vram;

library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;

ARCHITECTURE behavior OF pulse_widths_vram IS
BEGIN
    PROCESS (in0)
        VARIABLE prev_event : TIME := 0 ns;
    BEGIN
        IF prev_event /= 0 ns THEN
            IF in0 = '1' THEN
                ASSERT Now - prev_event > tlow
                    REPORT "Pulse width of '0' is less than " & time'image(tlow)
                    SEVERITY Warning;
            ELSIF in0 = '0' THEN
                ASSERT Now - prev_event > thigh
                    REPORT "Pulse width of '1' is less than " & time'image(thigh)
                    SEVERITY Warning;
            END IF;
        END IF;

        prev_event := Now;
    END PROCESS;

    out0 <= in0;
END behavior;

library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;

entity vram is
    generic ( wlength :integer := 512; width : integer := 8; sr_size : integer :=512;
              asize : integer :=9);
    port (
		a : in std_logic_vector(asize-1 downto 0);
        not_ras,not_cas,not_we,dsf,not_dtoe,sc,not_se : in std_logic;
		data : inout std_logic_vector(width-1 downto 0);
		qsf : out std_logic;	
		sio : inout std_logic_vector(width-1 downto 0)	
   );
end vram;

library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;


architecture operate_vram of vram is

type mem_type is array (wlength-1 downto 0,(sr_size/4)-1 downto 0) of integer;
type sr_type is array (sr_size-1 downto 0) of integer;
shared variable mem : mem_type;

    -- function evec2int (l: std_ulogic_vector)                      return natural;
    -- Translate unsigned vector into integer representation
    ---
       function evec2int (l: std_ulogic_vector)                      
             return integer is
          variable result: integer := 0;
           attribute synthesis_return of result:variable is "FEED_THROUGH" ;
       begin
          for t1 in l'range loop
             result := result * 2;
             if (l(t1) = '1' or l(t1) = 'H') then
                result := result + 1;
             end if;
          end loop;
          return result;
       end evec2int;


FUNCTION memout(mem : in mem_type; row,col : in integer) RETURN integer IS
variable temp : std_ulogic_vector(31 downto 0);
BEGIN
      temp := int2evec((mem(row,col/4)));
 CASE (col rem 4) IS
    WHEN 0 =>
      return evec2int(temp(7 downto 0));
    WHEN 1 =>
      return evec2int(temp(15 downto 8));
    WHEN 2 =>
      return evec2int(temp(23 downto 16));
    WHEN 3 =>
      return evec2int(temp(31 downto 24));
    WHEN OTHERS =>
 END CASE;      
END memout;

procedure memin(mem : inout mem_type; row,col : in integer; input : in integer) is
variable temp : std_ulogic_vector(31 downto 0);
begin
      temp := int2evec((mem(row,col/4)));
  
 CASE (col rem 4) IS
    WHEN 0 =>
      temp(7 downto 0) := int2evec(input)(7 downto 0);
    WHEN 1 =>
      temp(15 downto 8) := int2evec(input)(7 downto 0);
    WHEN 2 =>
      temp(23 downto 16) := int2evec(input)(7 downto 0);
    WHEN 3 =>
      temp(31 downto 24) := int2evec(input)(7 downto 0);
    WHEN OTHERS =>
 END CASE;
      mem(row,col/4) := evec2int(temp);   
  
end memin;

shared variable sr : sr_type;
shared variable sptr : integer :=0;
shared variable next_sptr : integer :=0;

signal mask_reg,color_reg : std_logic_vector(width-1 downto 0);

signal block_reg : std_logic_vector(3 downto 0);

signal mode : std_logic_vector(3 downto 0);
signal cmode : std_logic_vector(2 downto 0);

signal mask_mode : std_logic :='0'; -- 0, new mask mode, 1 old mask mode
signal shift_mode : std_logic :='0'; -- 0=read, 1=write
signal split_mode : std_logic :='0'; -- 0-reg, 1=split mode

shared variable row_a,col_a : integer;

signal qsft1,qsft2,qsfsel,qsfx : std_logic :='0';

-- Component instantiated by Turbo autoplace on 12/08/95 at 15:49:03
COMPONENT pulse_widths_vram
    GENERIC (
        tlow: TIME := 0 ns;
        thigh: TIME := 0 ns);
    PORT (
        in0 : IN std_logic;
        out0: OUT std_logic);
END COMPONENT;

SIGNAL  tnot_ras,tnot_cas,tsc : Std_logic; -- temps for pulse width checks



-- values for -70 parts
constant tASR : time := 0 ns;
constant tRAH : time := 10 ns;
constant tASC : time := 0 ns;
constant tCAH : time := 10 ns;
constant tCAC : time := 20 ns;
constant tOFC : time := 0 ns;  -- 0 min, 15 max
constant tSCA : time := 17 ns;
constant tSEZ : time := 0 ns;  -- 0 min 15 max
constant tPD : time := 20 ns;  
constant tDQR : time := 40 ns; -- 0 min 40 max
constant tWCS : time := 0 ns;
constant tWCH : time := 12 ns;
constant tDHS : time := 0 ns;
constant tDHH : time := 12 ns;
constant tFRS : time := 0 ns;
constant tFRH : time := 10 ns;
constant tRAS : time := 70 ns;
constant tRP : time := 50 ns;
constant tCAS : time := 10 ns;	-- for EDO hyper-page mode
constant tCP : time := 10 ns;
constant tSCL : time := 5 ns;
constant tSCH : time := 5 ns;

signal probe : time;
 

begin




-- not_ras to a setup and hold checks
assert1: for i in 0 to asize-1 generate
x : ASSERT not((not_ras='0' and not not_ras'stable) and (not a(i)'stable(tASC))) 
        REPORT "setup violation not_ras to a"
        SEVERITY warning;

y : ASSERT not((not_ras='0' and a(i)'event) and (not not_ras'stable(tCAH))) 
        REPORT "hold violation not_ras to a"
        SEVERITY warning;

end generate;

-- not_cas to a setup and hold checks
assert2: for i in 0 to asize-1 generate
x : ASSERT not((not_cas='0' and not not_cas'stable) and (not a(i)'stable(tASR))) 
        REPORT "setup violation not_cas to a"
        SEVERITY warning;

y : ASSERT not((not_cas='0' and a(i)'event) and (not not_cas'stable(tRAH))) 
        REPORT "hold violation not_cas to a"
        SEVERITY warning;

end generate;

-- not_cas to not_we setup and hold checks
g : ASSERT not((not_cas='0' and not not_cas'stable) and (not not_we'stable(tWCS))) 
        REPORT "setup violation not_cas to not_we"
        SEVERITY warning;

h : ASSERT not((not_cas='0' and not_we'event) and (not not_cas'stable(tWCH))) 
        REPORT "hold violation not_cas to not_we"
        SEVERITY warning;
-- not_ras to not_dtoe setup and hold checks
i : ASSERT not((not_ras='0' and not not_ras'stable) and (not not_dtoe'stable(tDHS))) 
        REPORT "setup violation not_ras to not_dtoe"
        SEVERITY warning;

k : ASSERT not((not_ras='0' and not_dtoe'event) and (not not_ras'stable(tDHH))) 
        REPORT "hold violation not_ras to not_dtoe"
        SEVERITY warning;
-- not_ras to dsf setup and hold checks
l : ASSERT not((not_ras='0' and not not_ras'stable) and (not dsf'stable(tFRS))) 
        REPORT "setup violation not_ras to dsf"
        SEVERITY warning;

m : ASSERT not((not_ras='0' and dsf'event) and (not not_ras'stable(tFRH))) 
        REPORT "hold violation not_ras to dsf"
        SEVERITY warning;

-- pulse width minimums tests
pnot_ras: pulse_widths_vram 
    GENERIC MAP (tlow => tRAS, thigh => tRP)
    PORT MAP (in0 => not_ras, out0 => tnot_ras);
pnot_cas: pulse_widths_vram 
    GENERIC MAP (tlow => tCAS, thigh => tCP)
    PORT MAP (in0 => not_cas, out0 => tnot_cas);
psc: pulse_widths_vram 
    GENERIC MAP (tlow => tSCL, thigh => tSCH)
    PORT MAP (in0 => sc, out0 => tsc);


proc_cas_ras: process
begin
    wait until (falling_edge(not_cas) and (not_ras='1'));
-- cas before ras operation
      cmode(2) <= not_dtoe;
      cmode(1) <= not_we;
      cmode(0) <= dsf;
      wait until falling_edge(not_ras);
	    CASE cmode IS
	    WHEN "001" | "101" => -- CBRS - CBR refresh/STOP (no reset)
          REPORT "CBRS - CBR refresh/STOP (no reset) - not supported"
          SEVERITY failure;
	    WHEN "011" | "111" => -- CBRn - CBR refresh (no reset)
	    WHEN "010" | "110" => -- CBRS - CBR refresh (option reset)
		  mask_mode <= '0'; -- mask mode is use new mask
	    WHEN OTHERS => -- X00 is undefined
          REPORT "not_dtoe=X not_we=0 dsf=0 - cas before ras - illegal mode"
          SEVERITY failure;
	    END CASE;
end process proc_cas_ras;

proc_ras_cas: process 
variable temp : std_logic_vector(width-1 downto 0);
variable temp_col_a : integer;
variable tempc : integer;
variable temp_col_sig : std_logic_vector(31 downto 0);
variable lowv,highv : integer;
begin
    wait until (falling_edge(not_ras) and (not_cas='1'));
-- ras before cas operation
-- capture mode
      mode(3) <= not_dtoe;
      mode(2) <= not_we;
      mode(1) <= dsf;
      row_a := evec2int(a);
	  if (mask_mode='0') then
	    mask_reg <= data;
	  end if;
      wait until falling_edge(not_cas) or rising_edge(not_ras);
	  if (not_cas='0') then
        mode(0) <= dsf;
        col_a := evec2int(a);
		block_reg <= data(3 downto 0);
        wait for 1 ns;
	    CASE mode IS
	    WHEN "0000" | "0001" => -- mwt masked write transfer
		  wait until (rising_edge(not_dtoe) or rising_edge(not_ras));
		  for i in 0 to (sr_size-1) loop
			temp := (int2evec(sr(i))(width-1 downto 0) and mask_reg) or (int2evec(memout(mem,row_a,i))(width-1 downto 0) and not mask_reg);
            memin(mem,row_a,i,evec2int(temp));
		  end loop;
		  sptr := col_a;
          next_sptr := 0;
		  shift_mode <= '1';
		  temp_col_sig := int2evec(col_a);
          qsft1 <= temp_col_sig(8) after tDQR;
		  qsfx <= '1';
		  wait for 1 ns;
		  qsfx <= '0';
	    WHEN "0010" | "0011" => -- mswt masked split write transfer
		-- transfer from inactive side of serial port
		  if (sptr >= (sr_size/2)) then
		    lowv := 0;
			highv := (sr_size/2)-1;
		  else
		    lowv := (sr_size/2);
			highv := sr_size-1;
		  end if;
		  for i in lowv to highv loop
			temp := (int2evec(sr(i))(width-1 downto 0) and mask_reg) or (int2evec(memout(mem,row_a,i))(width-1 downto 0) and not mask_reg);
            memin(mem,row_a,i,evec2int(temp));
		  end loop;
	    WHEN "0100" | "0101" => -- rt read transfer
		  wait until (rising_edge(not_dtoe) or rising_edge(not_ras));
		  for i in 0 to (sr_size-1) loop
		    sr(i) := memout(mem,row_a,i);
		  end loop;
		  sptr := col_a;
          next_sptr := 0;
		  shift_mode <= '0';
		  temp_col_sig := int2evec(col_a);
          qsft1 <= temp_col_sig(8) after tDQR;
		  qsfx <= '1';
		  wait for 1 ns;
		  qsfx <= '0';
	    WHEN "0110" | "0111" => -- srt split read transfer
          tempc := (col_a/(sr_size/2))*(sr_size/2); -- column offset
		  for i in 0 to ((sr_size/2)-1) loop
		    sr(i+tempc) := memout(mem,row_a,i+tempc);
		  end loop;
		  next_sptr := col_a;
	    WHEN "1000" => -- rwm read write (new/old mask)
		  while (not_ras='0') loop
            if (not_we='1') and (not_dtoe='0') then -- read
              data <= int2evec(memout(mem,row_a,col_a))(width-1 downto 0) after tCAC;
            elsif (not_we='0') then -- write
			  temp := (data and mask_reg) or (int2evec(memout(mem,row_a,col_a))(width-1 downto 0) and not mask_reg);
              memin(mem,row_a,col_a,evec2int(temp));
            end if;
            wait until (falling_edge(not_cas) or (not_ras='1') or (not_dtoe'event) or (falling_edge(not_we)));
            if (((not_dtoe='1') and (not_we='1')) or (not_ras='1')) then
              for i in 0 to width-1 loop
                data(i) <= 'Z' after tOFC;
              end loop;
			end if;
			if (falling_edge(not_cas)) then
              col_a := evec2int(a);
			end if;
            wait for 1 ns;
		  end loop;
	    WHEN "1001" => -- bwm block write (new/old mask)
		  while (not_ras='0') loop
            if (not_we='1') and (not_dtoe='0') then -- read
              data <= int2evec(memout(mem,row_a,col_a))(width-1 downto 0) after tCAC;
            elsif (not_we='0') then -- write
			  for i in 0 to 3 loop
			    if (block_reg(i)='1') then -- if enabled
				  temp_col_a :=	(col_a/4)*4 + i;
			      temp := (data and color_reg) or (int2evec(memout(mem,row_a,temp_col_a))(width-1 downto 0) and not color_reg);
                  memin(mem,row_a,temp_col_a,evec2int(temp));
				end if;
			  end loop;
            end if;
            wait until (falling_edge(not_cas) or (not_ras='1') or (not_dtoe'event) or (falling_edge(not_we)));
            if (((not_dtoe='1') and (not_we='1')) or (not_ras='1')) then
              for i in 0 to width-1 loop
                data(i) <= 'Z' after tOFC;
              end loop;
			end if;
			if (falling_edge(not_cas)) then
              col_a := evec2int(a);
			end if;
			if (falling_edge(not_we)) then
		      block_reg <= data(3 downto 0);
			end if;
            wait for 1 ns;
		  end loop;
	    WHEN "1010" | "1011" => -- fwm flash write (new/old mask)
		  while (not_ras='0') loop
            if (not_we='1') and (not_dtoe='0') then -- read
              data <= int2evec(memout(mem,row_a,col_a))(width-1 downto 0) after tCAC;
            elsif (not_we='0') then -- write
			  for i in 0 to (sr_size-1) loop
                memin(mem,row_a,i,evec2int(mask_reg));
			  end loop;
            end if;
            wait until (falling_edge(not_cas) or (not_ras='1') or (not_dtoe'event) or (falling_edge(not_we)));
            if (((not_dtoe='1') and (not_we='1')) or (not_ras='1')) then
              for i in 0 to width-1 loop
                data(i) <= 'Z' after tOFC;
              end loop;
			end if;
			if (falling_edge(not_cas)) then
              col_a := evec2int(a);
			end if;
            wait for 1 ns;
		  end loop;
	    WHEN "1100" => -- rw read write (no mask)
		  while (not_ras='0') loop
            if (not_we='1') and (not_dtoe='0') then -- read
              data <= int2evec(memout(mem,row_a,col_a))(width-1 downto 0) after tCAC;
            elsif (not_we='0') then -- write
              memin(mem,row_a,col_a,evec2int(data));
            end if;
            wait until (falling_edge(not_cas) or (not_ras='1') or (not_dtoe'event) or (falling_edge(not_we)));
            if (((not_dtoe='1') and (not_we='1')) or (not_ras='1')) then
              for i in 0 to width-1 loop
                data(i) <= 'Z' after tOFC;
              end loop;
			end if;
			if (falling_edge(not_cas)) then
              col_a := evec2int(a);
			end if;
            wait for 1 ns;
		  end loop;
	    WHEN "1101" => -- bw block write (no mask)
		  while (not_ras='0') loop
            if (not_we='1') and (not_dtoe='0') then -- read
              data <= int2evec(memout(mem,row_a,col_a))(width-1 downto 0) after tCAC;
            elsif (not_we='0') then -- write
			  for i in 0 to 3 loop
			    if (block_reg(i)='1') then -- if enabled
				  temp_col_a :=	(col_a/4)*4 + i;
                  memin(mem,row_a,temp_col_a,evec2int(color_reg));
				end if;
			  end loop;
            end if;
            wait until (falling_edge(not_cas) or (not_ras='1') or (not_dtoe'event) or (falling_edge(not_we)));
            if (((not_dtoe='1') and (not_we='1')) or (not_ras='1')) then
              for i in 0 to width-1 loop
                data(i) <= 'Z' after tOFC;
              end loop;
			end if;
			if (falling_edge(not_cas)) then
              col_a := evec2int(a);
			end if;
			if (falling_edge(not_we)) then
		      block_reg <= data(3 downto 0);
			end if;
            wait for 1 ns;
		  end loop;
	    WHEN "1110" => -- lmr load (old) mask register set cycle
		  while (not_ras='0') loop
            if (not_we='1') and (not_dtoe='0') then -- read -this is not shown in timing diagrams 
              data <= mask_reg after tCAC; --? tCAC?
            elsif (not_we='0') then -- write
              mask_reg <= data;
			  mask_mode <= '1';
            end if;
            wait until (falling_edge(not_cas) or (not_ras='1') or (not_dtoe'event) or (falling_edge(not_we)));
            if (((not_dtoe='1') and (not_we='1')) or (not_ras='1')) then
              for i in 0 to width-1 loop
                data(i) <= 'Z' after tOFC;
              end loop;
			end if;
            wait for 1 ns;
		  end loop;
	    WHEN "1111" => -- lcr load color reg
		  while (not_ras='0') loop
            if (not_we='1') and (not_dtoe='0') then -- read -this is not shown in timing diagrams 
              data <= color_reg after tCAC; --? tCAC?
            elsif (not_we='0') then -- write
              color_reg <= data;
            end if;
            wait until (falling_edge(not_cas) or (not_ras='1') or (not_dtoe'event) or (falling_edge(not_we)));
            if (((not_dtoe='1') and (not_we='1')) or (not_ras='1')) then
              for i in 0 to width-1 loop
                data(i) <= 'Z' after tOFC;
              end loop;
			end if;
            wait for 1 ns;
		  end loop;
	    WHEN OTHERS =>
	    END CASE;
	  end if;
end process proc_ras_cas;

proc_sc: process
begin
  wait until rising_edge(sc) or rising_edge(qsfx);
  if rising_edge(qsfx) then
    qsfsel <= '0' after tDQR;
  else
    if (split_mode='1') then -- split mode
      if ((sptr rem (sr_size/2))=(sr_size/2)) then
        sptr := next_sptr;
        next_sptr := (next_sptr+(sr_size/2)) rem sr_size;
      end if;
    end if;
    if (sptr=((sr_size/2)-2)) then
      qsft2 <= '1' after tPD;
      qsfsel <= '1' after tPD;
    end if;
    if (sptr=(sr_size-2)) then
      qsft2 <= '0' after tPD;
      qsfsel <= '1' after tPD;
    end if;
    if ((shift_mode='1') and (not_se='0')) then -- write
      sr(sptr) := evec2int(sio);
    end if;
    sptr := sptr+1;
	sptr := sptr rem sr_size;
  end if;
end process proc_sc;

proc_sout: process (shift_mode,not_se,sc)
begin
    if (shift_mode='0') then -- read
      if (not_se='0') then
        sio <= int2evec(sr(sptr))(width-1 downto 0) after tSCA;
      else
        for i in 0 to width-1 loop
          sio(i) <= 'Z' after tSEZ;
        end loop;
      end if;
    else -- write
      for i in 0 to width-1 loop
        sio(i) <= 'Z' after tSEZ;
      end loop;
    end if;
end process proc_sout;

proc_qsf: process (qsft1,qsft2,qsfsel)
begin
	qsf <= (not qsfsel and qsft1) or (qsfsel and qsft2);
end process proc_qsf;


  
end operate_vram;

